"""
* xrnn is a fast, lightweight Python machine learning framework that provides a simple user interface to build and
  train neural networks similar to Keras.
* xrnn only requires one third library dependency which is NumPy (almost likely you had it already installed).
* For more information, look at `layers.py` and `models.py` docstrings, and for a minimal working example of how to use
  the package to train a CNN on MNIST dataset, see `example.py`.
"""

__version__ = "1.1.1"
